# Authors: Florian Hofer
#
# License: BSD (3-clause)

"""Utilities for runtime and detection quality benchmarks."""

import time
from typing import Any, Dict, Iterator

import biosppy
import ecgdetectors
import heartpy
import heartpy.exceptions
import mne
import neurokit2
import numpy as np
import wfdb.processing

import sleepecg
from sleepecg.io.ecg_readers import ECGRecord


def reader_dispatch(data_dir: str, db_slug: str) -> Iterator[ECGRecord]:
    """
    Read ECG records from mitdb, ltdb or gudb.

    Parameters
    ----------
    data_dir : str
        Directory where all datasets are stored.
    db_slug : str
        Short identifier of a dataset, e.g. `'mitdb'`.

    Yields
    ------
    ECGRecord
        Each element in the generator is of type `ECGRecord` and contains
        the ECG signal (`.ecg`), sampling frequency (`.fs`), annotated beat
        indices (`.annotations`), `.lead`, and `.id`.
    """
    readers = {
        'gudb': sleepecg.io.read_gudb,
        'ltdb': sleepecg.io.read_ltdb,
        'mitdb': sleepecg.io.read_mitdb,
    }
    if db_slug not in readers:
        raise ValueError(f'Invalid db_slug: {db_slug}')
    yield from readers[db_slug](data_dir)


def detector_dispatch(ecg: np.ndarray, fs: float, detector: str) -> np.ndarray:
    """
    Provide a common interface for different heartbeat detectors.

    Parameters
    ----------
    ecg : np.ndarray
        ECG signal.
    fs : float
        Sampling frequency of the ECG signal in Hz.
    detector : str
        String identifier of the detector to be used.

    Returns
    -------
    np.ndarray
        Indices of detected heartbeats.
    """
    if detector == 'mne':
        detection = mne.preprocessing.ecg.qrs_detector(fs, ecg, verbose=False)
    elif detector == 'wfdb-xqrs':
        detection = wfdb.processing.xqrs_detect(ecg, fs, verbose=False)
    elif detector == 'pyecg-pantompkins':
        detection = ecgdetectors.Detectors(fs).pan_tompkins_detector(ecg)
    elif detector == 'biosppy-hamilton':
        detection = biosppy.signals.ecg.hamilton_segmenter(ecg, fs)[0]
    elif detector == 'heartpy':
        wd, m = heartpy.process(ecg, fs)
        detection = np.array(wd['peaklist'])[wd['binary_peaklist'].astype(bool)]
    elif detector == 'neurokit2-nk':
        clean_ecg = neurokit2.ecg.ecg_clean(ecg, int(fs), method='neurokit')
        detection = neurokit2.ecg.ecg_findpeaks(clean_ecg, int(fs), method='neurokit')['ECG_R_Peaks']  # noqa: E501
    elif detector == 'neurokit2-kalidas2017':
        clean_ecg = neurokit2.ecg.ecg_clean(ecg, int(fs), method='kalidas2017')
        detection = neurokit2.ecg.ecg_findpeaks(clean_ecg, int(fs), method='kalidas2017')['ECG_R_Peaks']  # noqa: E501
    elif detector == 'sleepecg-c':
        detection = sleepecg.detect_heartbeats(ecg, fs, backend='c')
    elif detector == 'sleepecg-numba':
        detection = sleepecg.detect_heartbeats(ecg, fs, backend='numba')
    elif detector == 'sleepecg-python':
        detection = sleepecg.detect_heartbeats(ecg, fs, backend='python')
    else:
        raise ValueError(f'Unknown QRS detector: {detector}')
    return np.asarray(detection)


def evaluate_single(
    record: ECGRecord,
    detector: str,
    signal_len: int,
    max_distance: float,
    calc_rri_similarity: bool,
) -> Dict[str, Any]:
    """
    Evaluate a heartbeat detector on a given annotated ECG record.

    Optionally, similarity measures between detected and annotated
    RR intervals can be calculated. As this requires interpolation, it may
    take some time for long signals.

    Parameters
    ----------
    record : ECGRecord
        As received from `reader_dispatch`.
    detector : str
        String identifier of the detector to be used.
    signal_len : int
        Length to which the signal should be sliced.
    max_distance : float
        Maximum temporal distance in seconds between detected and annotated
        beats to count as a successful detection.
    calc_rri_similarity : bool
        If `True`, calculate similarity measures between detected and
        annotated RR intervals (computationally expensive for long
        signals).

    Returns
    -------
    dict[str, Any]
        A dictionary containing evaluation results.
    """
    signal_len_samples = int(signal_len * record.fs*60)
    ecg = record.ecg[:signal_len_samples]
    annotation = record.annotation[record.annotation < signal_len_samples]
    fs = int(record.fs)

    try:
        start = time.perf_counter()
        detection = detector_dispatch(ecg, fs, detector)
        runtime = time.perf_counter() - start
        TP, FP, FN = sleepecg.compare_heartbeats(
            detection,
            annotation,
            int(max_distance * record.fs),
        )

        if calc_rri_similarity:
            pearsonr, spearmanr, rmse = sleepecg.rri_similarity(detection, annotation)

    except heartpy.exceptions.BadSignalWarning:
        runtime = np.nan
        TP = []
        FP = []
        FN = annotation

        if calc_rri_similarity:
            pearsonr = np.nan
            spearmanr = np.nan
            rmse = np.nan

    result = {
        'record_id': record.id,
        'lead': record.lead,
        'fs': record.fs,
        'num_samples': len(ecg),
        'detector': detector,
        'max_distance': max_distance,
        'runtime': runtime,
        'TP': len(TP),
        'FP': len(FP),
        'FN': len(FN),
    }
    if calc_rri_similarity:
        result.update({
            'pearsonr': pearsonr,
            'spearmanr': spearmanr,
            'rmse': rmse,
        })
    return result



# Authors: Florian Hofer
#
# License: BSD (3-clause)

"""Runtime and detection quality benchmarks for heartbeat detectors."""

import csv
import sys
import time
import warnings
from pathlib import Path

import numpy as np
import yaml
from tqdm import tqdm
from utils import detector_dispatch, evaluate_single, reader_dispatch

if len(sys.argv) != 2:
    print('Usage: python benchmark_detectors.py <benchmark>')
    exit()

with open('config.yml') as config_file:
    cfg = yaml.safe_load(config_file)

benchmark = sys.argv[1]
try:
    cfg = cfg[benchmark]
except KeyError:
    raise ValueError(f'Invalid benchmark: {benchmark!r}, available: {list(cfg)}.') from None


if cfg.get('suppress_warnings', False):
    warnings.filterwarnings('ignore')

outfile_dir = Path(cfg.get('outfile_dir', '.'))
outfile_dir.mkdir(parents=True, exist_ok=True)

db_slug = cfg['db_slug']
timestamp = time.strftime('%Y_%m_%d__%H_%M_%S')
csv_filepath = outfile_dir / f'{benchmark}__{db_slug}__{timestamp}.csv'
print(f'Storing results to {csv_filepath.resolve()}')

data_dir = Path(cfg.get('data_dir', '~/.sleepecg/datasets')).expanduser()
records = list(reader_dispatch(data_dir, db_slug))
print(f'Loaded {len(records)} records from {db_slug}.')


fieldnames = [
    'record_id', 'lead', 'fs', 'num_samples', 'detector', 'max_distance', 'runtime', 'TP',
    'FP', 'FN',
]
if cfg.get('calc_rri_similarity', False):
    fieldnames += ['pearsonr', 'spearmanr', 'rmse']

# Trigger jit compilation to make runtime benchmarks representative
if 'sleepecg-numba' in cfg['detectors']:
    detector_dispatch(records[0].ecg[:10*records[0].fs], records[0].fs, 'sleepecg-numba')


with open(csv_filepath, 'w', newline='') as csv_file:
    writer = csv.DictWriter(
        csv_file,
        fieldnames=fieldnames,
    )
    writer.writeheader()
    for signal_len in cfg['signal_lengths']:
        print(f'==== Signal length: {signal_len} minutes ====')
        for detector in cfg['detectors']:
            detector_results = []
            for record in tqdm(records, desc=detector, leave=False, disable=True):
                if len(record.ecg) < signal_len * record.fs*60:
                    continue

                detector_results.append(
                    evaluate_single(
                        record,
                        detector,
                        signal_len,
                        cfg.get('max_distance', 0.1),
                        cfg.get('calc_rri_similarity', False),
                    ),
                )

                writer.writerow(detector_results[-1])
                csv_file.flush()
            mean_runtime = np.nanmean([x['runtime'] for x in detector_results])
            signal_mins_per_runtime_sec = signal_len / mean_runtime
            print(f'  mean runtime: {mean_runtime:8.5f}s -> {signal_mins_per_runtime_sec:4.0f} minutes analyzed per second # {detector}')  # noqa: E501



"""A package for sleep stage classification using ECG data."""

from . import io
from .heartbeats import compare_heartbeats, detect_heartbeats, rri_similarity

__all__ = [
    'io',
    'detect_heartbeats',
    'compare_heartbeats',
    'rri_similarity',
]

__version__ = '0.3.0'



# Authors: Florian Hofer
#
# License: BSD (3-clause)

"""I/O- and download-related utilities."""

import hashlib
from pathlib import Path
from typing import Optional

import requests

__all__ = [
    'download_file',
]

_HASH_FUNCTIONS = {
    'md5': hashlib.md5,
    'sha256': hashlib.sha256,
}


def _calculate_checksum(filepath: Path, checksum_type: str) -> str:
    """
    Calculate the checksum for a file.

    Parameters
    ----------
    filepath : pathlib.Path
        Location of the file.
    checksum_type : {'md5', 'sha256'}
        Type of the checksum to calculate.

    Returns
    -------
    str
        The hexdigest of the checksum.
    """
    computed_hash = _HASH_FUNCTIONS[checksum_type]()
    with open(filepath, 'rb') as file:
        while True:
            chunk = file.read(8192)
            if not chunk:
                break
            computed_hash.update(chunk)
    return computed_hash.hexdigest()


def download_file(
    url: str,
    target_filepath: Path,
    checksum: Optional[str] = None,
    checksum_type: Optional[str] = None,
    verbose: bool = False,
):
    """
    Download a single file from `url` to `target_filepath`.

    In case `checksum` and `checksum_type` are provided, the downloaded
    file is verified. Raises a `RuntimeError` in case verification fails.

    Parameters
    ----------
    url : str
        URL to download from.
    target_filepath : pathlib.Path
        Location where the downloaded file will be stored.
    checksum : str, optional
        Checksum to verify the file against, by default `None`.
    checksum_type : str, optional
        Type of the checksum, by default `None`.
    verbose : bool, optional
        If `True`, output information during download. By default `False`.
    """
    if target_filepath.is_file():
        if checksum is not None and checksum_type is not None:
            calculated_checksum = _calculate_checksum(target_filepath, checksum_type)
            if calculated_checksum == checksum:
                if verbose:
                    print(f'Skipping {url}, already downloaded.')
                return

    target_filepath.parent.mkdir(parents=True, exist_ok=True)

    if verbose:
        print(f'Downloading {url}...')

    response = requests.get(url)
    response.raise_for_status()

    with open(target_filepath, 'wb') as file:
        file.write(response.content)

    if checksum is not None and checksum_type is not None:
        calculated_checksum = _calculate_checksum(target_filepath, checksum_type)
        if calculated_checksum != checksum:
            raise RuntimeError(
                f'Checksum mismatch for {target_filepath}:\n'
                f'    {checksum!r} (expected)\n'
                f'    {calculated_checksum!r} (calculated)',
            )



# Authors: Florian Hofer
#
# License: BSD (3-clause)

"""Plot results of runtime and detection quality benchmarks."""

import sys
from pathlib import Path

import numpy as np
import pandas as pd
import plotly.express as px

if len(sys.argv) != 2:
    print('Usage: python plot_benchmark_results.py <results>.csv')
    exit()

results_filepath = Path(sys.argv[1])
benchmark, db_slug, *_ = results_filepath.stem.split('__')
plot_filepath = results_filepath.with_suffix('.svg')
results = pd.read_csv(results_filepath).sort_values('detector')

if benchmark == 'runtime':
    fs = results['fs'][0]
    results['signal_len'] = results['num_samples'] / (results['fs'] * 3600)
    results = results.groupby(['detector', 'signal_len'], as_index=False).agg(
        mean_runtime=('runtime', 'mean'),
        std_runtime=('runtime', 'std'),
        n=('runtime', 'count'),
    )
    results['error'] = results['std_runtime'] / np.sqrt(results['n'])

    fig = px.line(
        results,
        x='signal_len',
        y='mean_runtime',
        error_y='error',
        color='detector',
        log_y=True,
        labels={
            'signal_len': 'signal length in hours',
            'mean_runtime': 'mean runtime in s',
        },
        title=f'Mean detector runtime for {db_slug.upper()} (fs={fs}Hz)',
        width=800,
        height=600,
        render_mode='svg',
    )
    fig.update_yaxes(rangemode='tozero')
    fig.write_image(plot_filepath)

elif benchmark == 'metrics':
    results['recall'] = results['TP'] / (results['TP'] + results['FN'])
    results['precision'] = results['TP'] / (results['TP'] + results['FP'])
    results['f1'] = 2 / (results['recall']**-1 + results['precision']**-1)
    fig = px.box(
        results.melt(id_vars=['detector'], value_vars=['recall', 'precision', 'f1']),
        color='detector',
        y='value',
        labels={'value': ''},
        facet_col='variable',
        width=1000,
        height=600,
        title=f'Metrics for {db_slug.upper()}',
    ).update_xaxes(range=[-0.4, 0.4]).update_yaxes(range=[0, 1.01])
    fig.for_each_annotation(lambda a: a.update(text=a.text.split('=')[1]))
    fig.write_image(plot_filepath, scale=1.5)

elif benchmark == 'rri_similarity':
    fig = px.box(
        results,
        y='pearsonr',
        color='detector',
        title=f'Pearson correlation coefficient for RRI timeseries from {db_slug.upper()}',
    ).update_yaxes(range=[-1.01, 1.01])
    fig.write_image(plot_filepath, scale=1.5)

else:
    raise ValueError(f'No plotting strategy defined for {results_filepath}.')



# %% Imports
import matplotlib.pyplot as plt
import numpy as np

from sleepecg import compare_heartbeats, detect_heartbeats
from sleepecg.io import read_mitdb

# %% Download and read data, run detector
record = list(read_mitdb(records_pattern='234'))[1]
detection = detect_heartbeats(record.ecg, record.fs)


# %% Evaluation and visualization
TP, FP, FN = compare_heartbeats(detection, record.annotation, int(record.fs/10))

t = np.arange(len(record.ecg)) / record.fs

fig, ax = plt.subplots(3, sharex=True, figsize=(10, 8))

ax[0].plot(t, record.ecg, color='k', zorder=1, label='ECG')
ax[0].scatter(
    record.annotation/record.fs, record.ecg[record.annotation],
    marker='o', color='g', s=50, zorder=2, label='annotation',
)
ax[0].set_ylabel('raw signal in mV')

ax[1].eventplot(
    detection / record.fs, linelength=0.5, linewidth=0.5,
    color='k', zorder=1, label='detection',
)
ax[1].scatter(
    FN/record.fs, np.ones_like(FN), marker='x',
    color='r', s=70, zorder=2, label='FN',
)
ax[1].scatter(
    FP/record.fs, np.ones_like(FP), marker='+',
    color='orange', s=70, zorder=2, label='FP',
)
ax[1].set_yticks([])
ax[1].set_ylabel('heartbeat events')

ax[2].plot(
    detection[1:] / record.fs, 60 / (np.diff(detection) / record.fs),
    label='heartrate in bpm',
)
ax[2].set_ylabel('beats per minute')
ax[2].set_xlabel('time in seconds')

for ax_ in ax.flat:
    ax_.legend(loc='upper right')
    ax_.grid(axis='x')

fig.suptitle(
    f'Record ID: {record.id}, lead: {record.lead}\n' +
    f'Recall: {len(TP) / (len(TP) + len(FN)):.2%}, ' +
    f'Precision: {len(TP) / (len(TP) + len(FP)):.2%}',
)

plt.show()



import numpy as np
from setuptools import Extension, setup

setup(
    ext_modules=[
        Extension(
            'sleepecg._heartbeat_detection',
            ['sleepecg/_heartbeat_detection.c'],
            include_dirs=[np.get_include()],
        ),
    ],
)



"""Functions for downloading and reading datasets."""

from .ecg_readers import read_gudb, read_ltdb, read_mitdb
from .nsrr import download_nsrr, set_nsrr_token

__all__ = [
    'read_gudb',
    'read_ltdb',
    'read_mitdb',
    'set_nsrr_token',
    'download_nsrr',
]
